# Window.py -- Pamhyr
# Copyright (C) 2023  INRAE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-

import logging

from tools import trace, timer

from View.Tools.PamhyrWindow import PamhyrWindow

from PyQt5.QtGui import (
    QKeySequence,
)

from PyQt5.QtCore import (
    Qt, QVariant, QAbstractTableModel,
    QCoreApplication, QModelIndex, pyqtSlot,
    QRect,
)

from PyQt5.QtWidgets import (
    QDialogButtonBox, QPushButton, QLineEdit,
    QFileDialog, QTableView, QAbstractItemView,
    QUndoStack, QShortcut, QAction, QItemDelegate,
    QComboBox, QVBoxLayout, QHeaderView, QTabWidget,
)

from View.InitialConditions.UndoCommand import (
    SetCommand, AddCommand, DelCommand,
    SortCommand, MoveCommand, PasteCommand,
    DuplicateCommand,
)

from View.InitialConditions.Table import InitialConditionTableModel, ComboBoxDelegate

from View.Tools.Plot.PamhyrCanvas import MplCanvas
from View.Tools.Plot.PamhyrToolbar import PamhyrPlotToolbar

from View.InitialConditions.PlotDKP import PlotDKP
from View.InitialConditions.PlotDischarge import PlotDischarge
from View.InitialConditions.translate import ICTranslate
from View.InitialConditions.DialogHeight import HeightDialog
from View.InitialConditions.DialogDischarge import DischargeDialog

_translate = QCoreApplication.translate

logger = logging.getLogger()

class InitialConditionsWindow(PamhyrWindow):
    _pamhyr_ui = "InitialConditions"
    _pamhyr_name = "Initial condition"

    def __init__(self, study=None, config=None, reach=None, parent=None):
        if reach is not None:
            self._reach = reach
        else:
            self._reach = study.river.current_reach()

        name = (
            self._pamhyr_name +
            " - " + study.name +
            " - " + self._reach.name
        )

        super(InitialConditionsWindow, self).__init__(
            title = name,
            study = study,
            config = config,
            trad = ICTranslate(),
            parent = parent
        )

        self._ics = study.river.initial_conditions.get(self._reach)

        self.setup_table()
        self.setup_plot()
        self.setup_connections()

        self.ui.setWindowTitle(self._title)

    def setup_table(self):
        table = self.find(QTableView, f"tableView")
        self._delegate_kp = ComboBoxDelegate(
            reach = self._reach,
            parent = self
        )

        self._table = InitialConditionTableModel(
            reach = self._reach,
            table_view = table,
            table_headers = self._trad.get_dict("table_headers"),
            editable_headers = ["kp", "discharge", "elevation", "height"],
            delegates = {"kp": self._delegate_kp},
            data = self._study,
            undo = self._undo_stack,
            trad = self._trad
        )

        table.setModel(self._table)
        table.setSelectionBehavior(QAbstractItemView.SelectRows)
        table.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
        table.setAlternatingRowColors(True)

    def setup_plot(self):
        self.canvas_1 = MplCanvas(width=5, height=4, dpi=100)
        self.canvas_1.setObjectName("canvas_1")
        self.toolbar_1 = PamhyrPlotToolbar(
            self.canvas_1, self
        )
        self.plot_layout_1 = self.find(QVBoxLayout, "verticalLayout_1")
        self.plot_layout_1.addWidget(self.toolbar_1)
        self.plot_layout_1.addWidget(self.canvas_1)

        self.plot_1 = PlotDKP(
            canvas = self.canvas_1,
            data = self._ics,
            trad = self._trad,
            toolbar = self.toolbar_1,
            parent = self
        )
        self.plot_1.draw()

        self.canvas_2 = MplCanvas(width=5, height=4, dpi=100)
        self.canvas_2.setObjectName("canvas_2")
        self.toolbar_2 = PamhyrPlotToolbar(
            self.canvas_2, self
        )
        self.plot_layout_2 = self.find(QVBoxLayout, "verticalLayout_2")
        self.plot_layout_2.addWidget(self.toolbar_2)
        self.plot_layout_2.addWidget(self.canvas_2)

        self.plot_2 = PlotDischarge(
            canvas = self.canvas_2,
            data = self._ics,
            trad = self._trad,
            toolbar = self.toolbar_2,
            parent = self
        )
        self.plot_2.draw()

    def setup_connections(self):
        self.find(QAction, "action_add").triggered.connect(self.add)
        self.find(QAction, "action_del").triggered.connect(self.delete)
        self.find(QAction, "action_sort").triggered.connect(self.sort)

        self.find(QPushButton, "pushButton_generate_1").clicked.connect(
            self.generate_growing_constante_height
        )

        self.find(QPushButton, "pushButton_generate_2").clicked.connect(
            self.generate_discharge
        )

        self._table.dataChanged.connect(self._update_plot)

    def index_selected_row(self):
        table = self.find(QTableView, f"tableView")
        return table.selectionModel()\
                    .selectedRows()[0]\
                    .row()

    def _update_plot(self):
        self.plot_1.draw()
        self.plot_2.draw()

    def index_selected_rows(self):
        table = self.find(QTableView, f"tableView")
        return list(
            # Delete duplicate
            set(
                map(
                    lambda i: i.row(),
                    table.selectedIndexes()
                )
            )
        )

    def add(self):
        rows = self.index_selected_rows()
        if len(self._ics) == 0 or len(rows) == 0:
            self._table.add(0)
        else:
            self._table.add(rows[0])

        self._update_plot()

    def delete(self):
        rows = self.index_selected_rows()
        if len(rows) == 0:
            return

        self._table.delete(rows)
        self._update_plot()

    def sort(self):
        self._table.sort(False)
        self._update_plot()

    def move_up(self):
        row = self.index_selected_row()
        self._table.move_up(row)
        self._update_plot()

    def move_down(self):
        row = self.index_selected_row()
        self._table.move_down(row)
        self._update_plot()

    def _copy(self):
        logger.info("TODO: copy")
        self._update_plot()

    def _paste(self):
        logger.info("TODO: paste")
        self._update_plot()

    def _undo(self):
        self._table.undo()
        self._update_plot()

    def _redo(self):
        self._table.redo()
        self._update_plot()

    def generate_growing_constante_height(self):
        dlg = HeightDialog(parent=self)
        if dlg.exec():
            value = dlg.value
            self._table.generate("growing", value)
            self._update_plot()

    def generate_discharge(self):
        dlg = DischargeDialog(parent=self)
        if dlg.exec():
            value = dlg.value
            self._table.generate("discharge", value)
            self._update_plot()
